/* Project based on the example framework ncs-spi-master-slave-example-master found here:
 * https://github.com/too1/ncs-spi-master-slave-example
 * Copyright (c) 2016 Intel Corporation
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr.h>
#include <device.h>
#include <devicetree.h>
#include <drivers/gpio.h>
#include <drivers/spi.h>
#include <stdlib.h>
#include<string.h>
#include<stdio.h>
#include<zephyr/random/rand32.h>

#define MY_SPI_MASTER DT_NODELABEL(my_spi_master)

// SPI master functionality
const struct device *spi_dev;
static struct k_poll_signal spi_done_sig = K_POLL_SIGNAL_INITIALIZER(spi_done_sig);

struct spi_cs_control spim_cs = {
	.gpio = SPI_CS_GPIOS_DT_SPEC_GET(DT_NODELABEL(reg_my_spi_master)),
	.delay = 0,
};

static void spi_init(void)
{
	spi_dev = DEVICE_DT_GET(MY_SPI_MASTER);
	if(device_is_ready(spi_dev) != 0) {
		printk("SPI master device not ready!\n");
	}
	if(device_is_ready(spim_cs.gpio.port) != 0){
		printk("SPI master chip select device not ready!\n");
	}
}

static const struct spi_config spi_cfg = {
	.operation = SPI_WORD_SET(8) | SPI_TRANSFER_MSB,   	// SPI-mode 0
	.frequency = 4000000,								// 4 MHZ
	.slave = 0,											// Master
	.cs = &spim_cs,										// Set Chip-Select pin
};

struct LINE{
	/*  Defining the line representation used in 
		the general framwork of co-operating robots project*/

	uint32_t x1;
	uint32_t y1;
	uint32_t x2;
	uint32_t y2;
	uint32_t theta;
};


static struct LINE spi_msg(char cap_command[20], int* num_lines, bool *waiting, int *idx)
{
	/*
	The spi_msg function handles the handles the spi-communication with the openMV M7,
	including the initialization of the buffers to be sent and received as well as the
	processing and output of the data received. It returns a line if received and
	sets the control variables num_lines, waiting and idx.
	*/

	// ----------------- Initialization -----------------
	char tx_buffer[30];
	memcpy(tx_buffer, cap_command, 30);					// Cast and transfer CAP command to local buffer
	static char rx_buffer[30];							// Cast buffer for received information

	const struct spi_buf tx_buf = {
		.buf = tx_buffer,
		.len = sizeof(tx_buffer)
	};
	const struct spi_buf_set tx = {
		.buffers = &tx_buf,
		.count = 1
	};

	struct spi_buf rx_buf = {
		.buf = rx_buffer,
		.len = sizeof(rx_buffer),
	};
	const struct spi_buf_set rx = {
		.buffers = &rx_buf,
		.count = 1
	};
	

	if(*waiting == false && *num_lines == 0){			// If CAP command is being sent
		printk("SPI TX: %s" , tx_buffer);
		printk("\n");
	}

	k_poll_signal_reset(&spi_done_sig);
	
	// ----------------- SPI transaction -----------------

	int error = spi_transceive_async(spi_dev, &spi_cfg, &tx, &rx, &spi_done_sig);

	if(error != 0){
		printk("SPI transceive error: %i\n", error);
		return;
	}

	int spi_signaled, spi_result;
	do{
		k_poll_signal_check(&spi_done_sig, &spi_signaled, &spi_result);
	} while(spi_signaled == 0);


	// ----------------- Data processing -----------------
	if(rx_buffer[0]== 0x4c && rx_buffer[1]== 0x49 && rx_buffer[2] == 0x4e){
		/*If LIN chars recevied, extract number of lines being sent*/
		int num_char = atoi(&rx_buffer[4]);
		char hold[3] = {0};
		int i = 0;
		while(num_char != 0){
			hold[i] = rx_buffer[6+i];
			--num_char;
			++i;
		}
		*num_lines = atoi(hold)+1;
		printk("NUMER OF LINES: %i  \n", *num_lines);
		*waiting = false;
	}

	if(rx_buffer[0]== 0x52 && rx_buffer[1]== 0x45 && rx_buffer[2] == 0x43 && *waiting == false){
		printk("Waiting for data\n");
		*waiting = true;
		++*idx;
	}
	struct LINE l = {0,0,0,0,0};

	if(*num_lines > 0){
		printk("SPI RX: %s \n", rx_buffer);
		--*num_lines;

		int i = 1;
		int hold_val[5];
		char *p;
		p = strtok(rx_buffer, ",");
		hold_val[0] = atoi(p);
		while(p != NULL){
			p = strtok(NULL, ",");
			if(p != NULL){
				hold_val[i] = atoi(p);
				++i;
			}
		}
		/*if(*num_lines > 0){
			l.x1 = hold_val[0]; l.y1 = hold_val[1]; l.x2 = hold_val[2]; l.y2 = hold_val[3]; l.theta = hold_val[4]; 
		}
		else{
			l.x1 = 0; l.y1 = 0; l.x2 =0; l.y2 =0; l.theta = 0;
		} */
	}

	// ---------------------------------------------------

	return l;
};

void get_pos(int *idx, char * arr){
	/* Returns a capture command string together with a positional measurement*/
	char positions[10][30]={
		{"CAP 100,100,150,225, \n"}, 
		{"CAP 100,100,150,180, \n"}, 
		{"CAP 32,100,2,45, \n"},
		{"CAP 120,400,22,45, \n"}, 
		{"CAP 12,100,534,45, \n"}, 
		{"CAP 200,53,2,45, \n"},
		{"CAP 521,15,2,45, \n"}, 
		{"CAP 223,124,12,45, \n"}, 
		{"CAP 65,100,124,45, \n"},
		{"CAP 113,4,2,45, \n"}};
	memcpy(arr, positions[*idx], 30);
	if(*idx == 9){*idx = 0;};

}

void main(void)
{
	/*
	The main function sends a capture command (CAP) and corresponding to the slave until a REC
	msg is received. When received it transitions to waiting mode, where it will send
	messages with filler data until a LIN msg is received. It will then receive and print
	out the amount of lines received. After printing it transitions into sending 
	CAP commands with updated position
	*/

	int num_lines = 0;
	char dont_care[30] = "FFFFFFFFFF\n";
	char cap_command[30] = "CAP ";
	int idx = 0;
	bool waiting = false;

	spi_init();
	printk("SPI communication started...\n");
	k_msleep(3000); 

	while (1) {
		if(waiting == false){
			get_pos(&idx, cap_command);
			spi_msg(cap_command, &num_lines, &waiting, &idx);
		}
		if(waiting == true){
			spi_msg(dont_care, &num_lines, &waiting, &idx);

			//printk("Received line: (x1: %i, y1: %i, x2: %i, y2: %i, theta:, %i \n", lin.x1, lin.y1, lin.x2, lin.y2, lin.theta);
		}
		k_msleep(100);
	}
}
