# Supporting functions

import sensor, image, time, gc, math


class MuLine:
    """ Mutable line object with extended functionality from the basic line type
        defined in openMV software (Non-mutable). Initiates with a openMV line
        object or a touple of the form (x1, y1, x2, y2) where the coordinates
        corresponds to start and stop point. """

    def __init__(self, line):

        if isinstance(line, tuple):     # Check if data is in touple form
            self.x_1 = line[0]
            self.y_1 = line[1]
            self.x_2 = line[2]
            self.y_2 = line[3]
            self.theta = line[4]
            self.rho = 0
            self.l = 0
        else:                           # Data is openMv object
            if line.x1() > line.x2():   # Flip orientation so line starts from the left
                self.x_1 = line.x2()
                self.x_2 = line.x1()
                self.y_1 = line.y2()
                self.y_2 = line.y1()
            else:
                self.x_1 = line.x1()
                self.x_2 = line.x2()
                self.y_1 = line.y1()
                self.y_2 = line.y2()
            self.theta = line.theta()
            self.rho = line.rho()
            self.l = line.length()
        if self.theta > 170:
            self.theta = 0

    def print(self):
        print("x1:", self.x_1, "y1:", self.y_1,
              "x2:", self.x_2, "y2:", self.y_2,
              "theta:", self.theta, "rho:", self.rho,
              "length:", self.l)

    def get_cord(self):

        return self.x_1, self.y_1, self.x_2, self.y_2

    def length(self):
        l = math.sqrt((self.x_2-self.x_1)**2+(self.y_2-self.y_1)**2)
        return l


def merge_lines(params, lines):
    """Iteratively matches and merges the passed lines until no
        more matches can be found. """
    line_arr = []                   # Array to store merged lines
    vertical = False                # Line is close to vertical
    extremal_points = None          # Stores start/end points if vertical = True
    fin_merge = False               # True when no more matches can be found
    match = False

    for l in lines:
        line_arr.append(MuLine(l))                      # Casts openMV lines as MuLine
    line_arr = quicksort(line_arr, 0, len(line_arr)-1)  # Sort by x-coordinate

    del lines
    gc.collect()

    print("Merging...")
    while fin_merge == False:

        merged_lines = []
        org_size = len(line_arr)

        for i in range(len(line_arr)):
            """ Iterates through list of lines. If a match is found the lines are
                merged and the result appended to the list used in the next iteration.
                The matched line is deleted from the working list so that it cannot be
                matched again """

            if i >= len(line_arr):              # original index out of bounds
                break

            line_seg = line_arr[i]              # Current line segment to match list with

            for j in range(i+1,len(line_arr)):  # For all elements after the current line

                match, vertical, extremal_points = is_match(params, line_arr[j], line_seg)

                if match == True:
                    merged_line = merge(vertical, [line_seg, line_arr[j]], extremal_points)
                    #print("MATCH")
                    #line_seg.print()
                    #line_arr[j].print()
                    line_seg = MuLine(merged_line)
                    #line_seg.print()
                    #print("\n")
                    del line_arr[j], merged_line

                if (j + 1) >= len(line_arr):    # original index out of bounds
                    break

            merged_lines.append(line_seg)

        line_arr = merged_lines                 # New working list

        if org_size == len(line_arr):           # No new matches
            fin_merge = True

    del merged_lines, org_size, match, vertical, fin_merge
    gc.collect()
    print("Merge complete")

    return line_arr


def is_match(params, l_1, l_2):
    """ Checks if two lines are matching. First checks if angle
        is corresponding, then checks if the lines x-displacement
        corresponds to a vertical line. If no, handle through
        extrapolation. If yes handle through end-point matching. """

    match = False
    vertical = False
    extremal_points = None
    theta_bound = [l_2.theta - params[5], l_2.theta + params[5]] # Max angle mismatch
    l_1_x_dist = l_1.x_2-l_1.x_1                                 # line 1 x-displacement
    l_2_x_dist = l_2.x_2-l_2.x_1                                 # line 2 x-displacement

    if l_1.theta > theta_bound[0] and l_1.theta < theta_bound[1]:
        if l_1_x_dist > 5 or l_2_x_dist > 5:                    # Line is not close to vertical
             match = regular_match(params, l_2, l_1)             # Extrapolation matching

        if match == False and (l_1_x_dist < 15 and l_2_x_dist < 15):       # Line close to vertical
            #print("Tying vertical:")
            #l_1.print()
            #l_2.print()
            #print("\n")
            match, vertical, extremal_points = vertical_match(l_1, l_2) # End-point matching

    del theta_bound, l_1_x_dist, l_2_x_dist
    gc.collect()

    return match, vertical, extremal_points


def vertical_match(l_1, l_2):
    """ Determines if two lines are corresponding matching by checking the four cases:

        1) l_1 startpoint matches l_2 startpoint
        2) l_1 startpoint matches l_2 endpoint
        3) l_1 endpoint matches l_2 startpoint
        4) l_1 endpoint matches l_2 endpoint

        Returns match if any case is within bounds, as well as the points corresponding
        to the longest y-axis displacement."""

    match = False
    vertical = False
    extremal_points = [(0,0),(0,0)]

    case_1 = [abs(l_2.x_1-l_1.x_1), abs(l_2.y_1-l_1.y_1), (l_1.x_2, l_1.y_2) ,(l_2.x_2, l_2.y_2)]
    case_2 = [abs(l_2.x_2-l_1.x_1), abs(l_2.y_2-l_1.y_1), (l_1.x_2, l_1.y_2) ,(l_2.x_1, l_2.y_1)]
    case_3 = [abs(l_2.x_1-l_1.x_2), abs(l_2.y_1-l_1.y_2), (l_1.x_1, l_1.y_1) ,(l_2.x_2, l_2.y_2)]
    case_4 = [abs(l_2.x_2-l_1.x_2), abs(l_2.y_2-l_1.y_2), (l_1.x_1, l_1.y_1) ,(l_2.x_1, l_2.y_1)]

    cases = [case_1, case_2, case_3, case_4]

    for case in cases:
        if case[0] < 5 and case[1] < 25:    # End points within bound
            match = True
            vertical = True
            if abs(extremal_points[0][1] - extremal_points[1][1]) < abs(case[2][1] - case[3][1]):
                # Switch extremal points with the largest y-displacement
                extremal_points = [case[2],case[3]]

    del case_1, case_2, case_3, case_4, cases
    gc.collect()
    return match, vertical, extremal_points


def regular_match(params, l_1, l_2):
    """Looks for matches by extrapolating the first endpoint to the second startpoint.
        If the actual y-coordinate of the second startpoint is within a set bound
        of the  extrapolated y-coordinate a match is detected. """

    upper_b = 0
    lower_b = 0
    slope = 0
    temp = None
    match = False
    y_dist = l_2.y_1-l_1.y_2            # x-distance between l_1 endpoint and l_2 startpoint
    x_dist = l_2.x_1-l_1.x_2            # y-distance between l_1 endpoint and l_2 startpoint

    if abs(y_dist) < params[6] and abs(x_dist) < params[6]:         # Within user bounds
        try:

            slope = (l_1.y_2-l_1.y_1) / (l_1.x_2-l_1.x_1)           # Find l_1 slope
            upper_b = int(l_1.y_2 + (slope*x_dist)+ params[7])      # Extrapolate to l_2 start
            lower_b = int(l_1.y_2 + (slope*x_dist)- params[7])      # Extrapolate to l_2 start

            if upper_b < lower_b:   # Slope is negative
               temp = lower_b
               lower_b = upper_b
               upper_b = temp

            if l_2.y_1 > lower_b  and l_2.y_1 < upper_b:            # Within extrapolated bounds
                match = True
        except:
            # Divide by zero in slope, handle by vertical matching
            pass

    del upper_b, lower_b, y_dist, x_dist, slope, temp
    gc.collect()

    return match


def merge(vertical, lines, extremal_points):
    """ Merges two lines by finding the middle point of each line,
        creating a line segment between these points and extrapolating
        the end and start point. The extrapolation is done by finding
        the slope of the new line and multiplying it with the difference
        between the original x coordinates and the middle point x-values. """

    m_1 = middle_point(lines[0])
    m_2 = middle_point(lines[1])
    start_point = (0,0)
    end_point = (0,0)
    slope = 0
    theta = 0

    if vertical == False:
        try:
            slope = (m_2[1]-m_1[1]) / (m_2[0]-m_1[0])   # Find slope of midpoint-line
            if abs(slope) > 15:                         # To steep, treat as vertical
                vertical = True
                _,_,extremal_points = vertical_match(lines[0], lines[1])

            if vertical == False:
                # Extrapolate line start and end
                start_point = (lines[0].x_1, int(m_1[1]-(slope*abs(m_1[0]-lines[0].x_1))))
                end_point = (lines[1].x_2, int(m_2[1]+(slope*abs(m_2[0]-lines[1].x_2))))
                theta = 90 + int((math.atan2(slope,1)*180) / math.pi)
        except:
            # Division by zero error if the averaged line points straight up, treat as vertical
            vertical = True
            _,_,extremal_points = vertical_match(lines[0], lines[1])

    elif vertical == True:
        start_point = extremal_points[0]
        end_point = extremal_points[1]
        theta = 0

    new_line = (start_point[0], start_point[1], end_point[0], end_point[1], theta)

    del m_1, m_2, start_point, end_point, theta, slope
    gc.collect()

    return new_line


def middle_point(line_seg):
    """ Find the middle point of a line """
    x_mid = (line_seg.x_1+line_seg.x_2) / 2
    y_mid = (line_seg.y_1+line_seg.y_2) / 2
    avg =(int(x_mid), int(y_mid))
    del x_mid, y_mid
    gc.collect()

    return avg


def calculate_distance(l, pos, image_size):
    real_line = MuLine((0,0,0,0,0))
    a = 0
    sensor_width = 3.984
    focal_length = 2.80
    wall_height = 30
    image_height = pos[2] - wall_height
    Iw = image_size[0]+24
    Ih = image_size[1]
    center_x = (image_size[0] / 2)+25
    center_y = (image_size[1] / 2)+55
    GSD = ((sensor_width * image_height) /(focal_length * Iw))*0.913

    real_line.l = GSD*l.length()

    start_p = [(GSD *(l.x_1 - center_x)), (GSD *(Ih-l.y_1 - center_y))]
    stop_p =  [(GSD *(l.x_2 - center_x)), (GSD *(Ih-l.y_2 - center_y))]

    real_line.x_1 = start_p[0] * math.cos(pos[3]) - start_p[1] * math.sin(pos[3])+ pos[0]
    real_line.y_1 = start_p[0] * math.sin(pos[3]) + start_p[1] * math.cos(pos[3])+ pos[1]
    real_line.x_2 = stop_p[0]  * math.cos(pos[3]) - stop_p[1]  * math.sin(pos[3])+ pos[0]
    real_line.y_2 = stop_p[0]  * math.sin(pos[3]) + stop_p[1]  * math.cos(pos[3])+ pos[1]
    print(start_p, stop_p)
    print(math.sin(pos[3]), math.cos(pos[3]))
    print(GSD, Ih, l.y_1, center_y)
    print(GSD, Ih, l.y_2, center_y)

    try:
        a = (real_line.y_2-real_line.y_1) / (real_line.x_2-real_line.x_1)
        real_line.theta = int((math.atan2(a,1)*180) / math.pi)
    except:
        real_line.theta =  0

    del sensor_width, focal_length, wall_height, image_height, Iw, Ih
    del center_x, center_y, GSD, start_p, stop_p, a
    gc.collect()

    return real_line


def quicksort(array, low, high):
    """Classic quicksort, sorts the x_1 coordinates
       of a list of lines """

    if low < high:
        pi = partition(array, low, high)
        quicksort(array, low, pi - 1)
        quicksort(array, pi + 1, high)

    return array


def partition(array, low, high):

    pivot = array[high].x_1

    i = low - 1

    for j in range(low, high):
        j_call = array[j].x_1

        if j_call <= pivot:

            i = i + 1

            (array[i], array[j]) = (array[j], array[i])

    (array[i + 1], array[high]) = (array[high], array[i + 1])
    del pivot, j_call, j,
    gc.collect()

    return i + 1







